from __future__ import annotations
from pydantic import BaseModel, StrictStr, StrictInt
from typing import Optional

class Message(BaseModel):
    host: Optional[StrictStr]
    port: Optional[StrictInt]

    def to_data(self):
        return self.dict(
            exclude={
                'host',
                'port'
            }
        )
    



import random
import inspect
from mercury_sync.connection.tcp.mercury_sync_tcp_connection import MercurySyncTCPConnection
from mercury_sync.connection.udp.mercury_sync_udp_connection import MercurySyncUDPConnection
from mercury_sync.models.message import Message
from typing import (
    Tuple, 
    Dict, 
    TypeVar, 
    Generic,
    AsyncIterable
)


T = TypeVar('T', bound=Message)
R = TypeVar('R', bound=Message)


class Service(Generic[T, R]):
    
    def __init__(
        self,
        host: str,
        port: int
    ) -> None:
        self.name = self.__class__.__name__
        self._instance_id = random.randint(0, 2**16)


        self._udp_connection = MercurySyncUDPConnection(
            host,
            port,
            self._instance_id
        )

        self._tcp_connection = MercurySyncTCPConnection(
            host,
            port + 1,
            self._instance_id
        )

        self._host_map: Dict[str, Tuple[str, int]] = {}    

        methods = inspect.getmembers(self, predicate=inspect.ismethod)

        reserved_methods = [
            'start',
            'connect',
            'send',
            'send_direct',
            'close'
        ]

        for _, method in methods:
            method_name = method.__name__

            not_internal = method_name.startswith('__') is False
            not_reserved = method_name not in reserved_methods
            is_server = hasattr(method, 'server_only')

            if not_internal and not_reserved and is_server:
                self._tcp_connection.events[method.__name__] = method
                self._udp_connection.events[method.__name__] = method

    def start(self):
        self._tcp_connection.connect()
        self._udp_connection.connect()

    async def connect(
        self,
        remote: T
    ):
        address = (remote.host, remote.port)
        self._host_map[remote.__class__.__name__] = address

        await self._tcp_connection.connect_client((
            remote.host, 
            remote.port + 1
        ))

    async def send(
        self, 
        event_name: str,
        message: T
    ) -> R:
        (host, port)  = self._host_map.get(message.__class__.__name__)
        address = (
            host,
            port
        )

        return await self._udp_connection.send(
            event_name,
            message.to_data(),
            address
        )
    
    async def send_direct(
        self,
        event_name: str,
        message: T
    ) -> R:
        (host, port)  = self._host_map.get(message.__class__.__name__)
        address = (
            host,
            port + 1
        )

        return await self._tcp_connection.send(
            event_name,
            message.to_data(),
            address
        )
    
    async def stream(
        self,
        event_name: str,
        message: T
    ) -> AsyncIterable[R]:
        (host, port)  = self._host_map.get(message.__class__.__name__)
        address = (
            host,
            port
        )

        async for response in self._udp_connection.stream(
            event_name,
            message.to_data(),
            address
        ):
            yield response

    async def stream_direct(
        self,
        event_name: str,
        message: T
    ) -> AsyncIterable[R]:
        (host, port)  = self._host_map.get(message.__class__.__name__)
        address = (
            host,
            port + 1
        )

        async for response in self._tcp_connection.stream(
            event_name,
            message.to_data(),
            address
        ):
            yield response
    
    async def close(self):
        await self._tcp_connection.close()



import functools
from mercury_sync.service import Service


def client(
    call_name: str, 
    direct: bool=False
):

    def wraps(func):

        func.client_only = True

        @functools.wraps(func)
        async def decorator(
            *args,
            **kwargs
        ):
            connection: Service = args[0]

            if direct:
                return await connection.send_direct(
                    call_name,
                    await func(*args, **kwargs)
                )

            else:
                return await connection.send(
                    call_name,
                    await func(*args, **kwargs)
                )
            
        return decorator
    
    return wraps




import asyncio
import pickle
import zlib
from collections import deque, defaultdict
from mercury_sync.snowflake.snowflake_generator import SnowflakeGenerator
from mercury_sync.models.message import Message
from typing import Tuple, Deque, Any, Dict, Coroutine, AsyncIterable
from mercury_sync.connection.tcp.protocols import (
    MercurySyncTCPClientProtocol,
    MercurySyncTCPServerProtocol
)


class MercurySyncTCPConnection:

    def __init__(
        self,
        host: str,
        port: int,
        instance_id: int
    ) -> None:

        self.id_generator = SnowflakeGenerator(instance_id)

        self.host = host
        self.port = port

        self.events: Dict[
            str,
            Coroutine
        ] = {}

        self._client_transports: Dict[str, asyncio.Transport] = {}
        self._server: asyncio.Server = None
        self._loop = asyncio.get_event_loop()
        self.queue: Dict[str, Deque[Tuple[str, int, float, Any]] ] = defaultdict(deque)
        self._waiters: Dict[str, Deque[asyncio.Future]] = defaultdict(deque)
        self._pending_responses = deque()
        self._sent_values = deque()
    
    def connect(self):
        server = self._loop.create_server(
            lambda: MercurySyncTCPServerProtocol(
                self.read
            ),
            host=self.host,
            port=self.port
        )

        
        self._server = self._loop.run_until_complete(server)

    async def connect_client(
        self,
        address: Tuple[str, int]
    ):

        host, port = address

        client_transport, _ = await self._loop.create_connection(
            lambda: MercurySyncTCPClientProtocol(
                self.read
            ),
            host=host,
            port=port
        )

        self._client_transports[address] = client_transport

        return client_transport


    async def send(
        self, 
        event_name: bytes,
        data: bytes, 
        address: Tuple[str, int]
    ) -> str:
        
        client_transport = self._client_transports.get(address)

        item = pickle.dumps((
            'request',
            next(self.id_generator),
            event_name,
            data,
            self.host,
            self.port
        ))

        compressed = zlib.compress(item)

        client_transport.write(compressed)

        waiter = self._loop.create_future()
        self._waiters[event_name].append(waiter)

        await waiter

        (
            _,
            shard_id,
            _,
            response_data,
            _, 
            _
        ) = self.queue[event_name].pop()

        return (
            shard_id,
            response_data
        )
    
    async def stream(
        self, 
        event_name: str,
        data: Any, 
        address: Tuple[str, int]
    ): 
        client_transport = self._client_transports.get(address)

        item = pickle.dumps((
            'stream',
            next(self.id_generator),
            event_name,
            data,
            self.host,
            self.port
        ))

        compressed = zlib.compress(item)

        client_transport.write(compressed)

        waiter = self._loop.create_future()
        self._waiters[event_name].append(waiter)

        await waiter

        for item in self.queue[event_name]:
            (
                _,
                shard_id,
                _,
                response_data,
                _, 
                _
            ) = item
         
            yield(
                shard_id,
                response_data
            )

        self.queue.clear()

    def read(
        self,
        data: bytes,
        transport: asyncio.Transport
    ):
    
        result: Tuple[
            str, 
            int, 
            float, 
            Any, 
            str, 
            int
        ] = pickle.loads(
            zlib.decompress(data)
        )

        (
            message_type, 
            shard_id, 
            event_name,
            payload, 
            incoming_host, 
            incoming_port
        ) = result

        if message_type == 'request':
            self._pending_responses.append(
                asyncio.create_task(
                    self._read(
                        event_name,
                        self.events.get(event_name)(
                            shard_id,
                            payload
                        ),
                        transport
                    )
                )
            )

        elif message_type == 'stream':

            self._pending_responses.append(
                asyncio.create_task(
                    self._read_iterator(
                        event_name,
                        self.events.get(event_name)(
                            shard_id,
                            payload
                        ),
                        transport
                    )
                )
            )

        else:
            print(result)
            self.queue[event_name].append((
                message_type, 
                shard_id,
                event_name,
                payload, 
                incoming_host,
                incoming_port
            ))

                
            event_waiter = self._waiters[event_name]

            if len(event_waiter) > 0:
                waiter = event_waiter.pop()
                waiter.set_result(None)

    async def _read(
        self,
        event_name: str,
        coroutine: Coroutine,
        transport: asyncio.Transport
    ):
        response: Message = await coroutine

        item = pickle.dumps(
            (
                'response', 
                next(self.id_generator),
                event_name,
                response.to_data(), 
                self.host,
                self.port
            )
        )

        compressed = zlib.compress(item)

        transport.write(compressed)

    async def _read_iterator(
        self,
        event_name: str,
        coroutine: AsyncIterable[Message],
        transport: asyncio.Transport       
    ):
        
        async for response in coroutine:
            item = pickle.dumps(
                (
                    'response', 
                    next(self.id_generator),
                    event_name,
                    response.to_data(), 
                    self.host,
                    self.port
                )
            )

            compressed = zlib.compress(item)
            transport.write(compressed)
            

    async def close(self):
        for transport in self._client_transports.values():
            transport.close()

        self._server.close()